import { ElementEvent, Shape, Path, isNil, CustomElement } from '@antv/g';

function _extends() {
  _extends = Object.assign ? Object.assign.bind() : function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };
  return _extends.apply(this, arguments);
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;

  _setPrototypeOf(subClass, superClass);
}

function _setPrototypeOf(o, p) {
  _setPrototypeOf = Object.setPrototypeOf ? Object.setPrototypeOf.bind() : function _setPrototypeOf(o, p) {
    o.__proto__ = p;
    return o;
  };
  return _setPrototypeOf(o, p);
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

/**
 * Common utilities
 * @module glMatrix
 */
var ARRAY_TYPE = typeof Float32Array !== 'undefined' ? Float32Array : Array;
if (!Math.hypot) Math.hypot = function () {
  var y = 0,
      i = arguments.length;

  while (i--) {
    y += arguments[i] * arguments[i];
  }

  return Math.sqrt(y);
};

/**
 * 3 Dimensional Vector
 * @module vec3
 */

/**
 * Creates a new, empty vec3
 *
 * @returns {vec3} a new 3D vector
 */

function create() {
  var out = new ARRAY_TYPE(3);

  if (ARRAY_TYPE != Float32Array) {
    out[0] = 0;
    out[1] = 0;
    out[2] = 0;
  }

  return out;
}
/**
 * Creates a new vec3 initialized with the given values
 *
 * @param {Number} x X component
 * @param {Number} y Y component
 * @param {Number} z Z component
 * @returns {vec3} a new 3D vector
 */

function fromValues(x, y, z) {
  var out = new ARRAY_TYPE(3);
  out[0] = x;
  out[1] = y;
  out[2] = z;
  return out;
}
/**
 * Subtracts vector b from vector a
 *
 * @param {vec3} out the receiving vector
 * @param {ReadonlyVec3} a the first operand
 * @param {ReadonlyVec3} b the second operand
 * @returns {vec3} out
 */

function subtract(out, a, b) {
  out[0] = a[0] - b[0];
  out[1] = a[1] - b[1];
  out[2] = a[2] - b[2];
  return out;
}
/**
 * Alias for {@link vec3.subtract}
 * @function
 */

var sub = subtract;
/**
 * Perform some operation over an array of vec3s.
 *
 * @param {Array} a the array of vectors to iterate over
 * @param {Number} stride Number of elements between the start of each vec3. If 0 assumes tightly packed
 * @param {Number} offset Number of elements to skip at the beginning of the array
 * @param {Number} count Number of vec3s to iterate over. If 0 iterates over entire array
 * @param {Function} fn Function to call for each vector in the array
 * @param {Object} [arg] additional argument to pass to fn
 * @returns {Array} a
 * @function
 */

var forEach = function () {
  var vec = create();
  return function (a, stride, offset, count, fn, arg) {
    var i, l;

    if (!stride) {
      stride = 3;
    }

    if (!offset) {
      offset = 0;
    }

    if (count) {
      l = Math.min(count * stride + offset, a.length);
    } else {
      l = a.length;
    }

    for (i = offset; i < l; i += stride) {
      vec[0] = a[i];
      vec[1] = a[i + 1];
      vec[2] = a[i + 2];
      fn(vec, vec, arg);
      a[i] = vec[0];
      a[i + 1] = vec[1];
      a[i + 2] = vec[2];
    }

    return a;
  };
}();

var _excluded = ["body", "startHead", "endHead", "startHeadOffset", "endHeadOffset"],
    _excluded2 = ["body", "startHead", "endHead", "startHeadOffset", "endHeadOffset"],
    _excluded3 = ["body", "startHead", "endHead", "startHeadOffset", "endHeadOffset"];
/**
 * support 3 types of arrow line:
 * 1. Line
 * 2. Polyline
 * 3. Path
 *
 * support 2 types of arrow head:
 * 1. default(Path)
 * 2. custom
 */

var Arrow = /*#__PURE__*/function (_CustomElement) {
  _inheritsLoose(Arrow, _CustomElement);

  function Arrow(config) {
    var _this;

    _this = _CustomElement.call(this, _extends({}, config, {
      type: Arrow.tag
    })) || this; // eslint-disable-next-line @typescript-eslint/no-unused-vars

    _this.body = void 0;
    _this.startHead = void 0;
    _this.endHead = void 0;
    _this.startHeadPosition = void 0;
    _this.startHeadRad = void 0;
    _this.endHeadPosition = void 0;
    _this.endHeadRad = void 0;

    var _this$attributes = _this.attributes,
        body = _this$attributes.body,
        startHead = _this$attributes.startHead,
        endHead = _this$attributes.endHead,
        startHeadOffset = _this$attributes.startHeadOffset,
        endHeadOffset = _this$attributes.endHeadOffset,
        rest = _objectWithoutPropertiesLoose(_this$attributes, _excluded);

    if (!body) {
      throw new Error("Arrow's body is required");
    } // append arrow body


    _this.body = body;

    _this.appendChild(_this.body);

    _this.handleBodyAttributeChanged(_this.body);

    if (startHead) {
      _this.appendArrowHead(_this.getArrowHeadType(startHead), true);
    }

    if (endHead) {
      _this.appendArrowHead(_this.getArrowHeadType(endHead), false);
    }

    _this.applyArrowStyle(rest, [_this.body, _this.startHead, _this.endHead]);

    return _this;
  }

  var _proto = Arrow.prototype;

  _proto.handleBodyAttributeChanged = function handleBodyAttributeChanged(body) {
    var _this2 = this;

    body.addEventListener(ElementEvent.ATTR_MODIFIED, function (e) {
      var attrName = e.attrName;

      if (attrName === 'x1' || attrName === 'y1') {
        if (_this2.startHead) {
          _this2.transformArrowHead(_this2.startHead, true);
        }
      } else if (attrName === 'x2' || attrName === 'y2') {
        if (_this2.endHead) {
          _this2.transformArrowHead(_this2.endHead, false);
        }
      } // const { nodeName } = body;
      // if (nodeName === Shape.LINE) {
      //   body
      // }

    });
  };

  _proto.getBody = function getBody() {
    return this.body;
  };

  _proto.getStartHead = function getStartHead() {
    return this.startHead;
  };

  _proto.getEndHead = function getEndHead() {
    return this.endHead;
  };

  _proto.attributeChangedCallback = function attributeChangedCallback(name, oldValue, newValue) {
    if (name === 'opacity' || name === 'strokeOpacity' || name === 'stroke' || name === 'lineWidth' || name === 'increasedLineWidthForHitTesting') {
      var _this$applyArrowStyle;

      this.applyArrowStyle((_this$applyArrowStyle = {}, _this$applyArrowStyle[name] = newValue, _this$applyArrowStyle), [this.body, this.startHead, this.endHead]);
    } else if (name === 'startHead' || name === 'endHead') {
      var isStart = name === 'startHead'; // delete existed arrow head first

      this.destroyArrowHead(isStart);

      if (newValue) {
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        var _this$attributes2 = this.attributes,
            body = _this$attributes2.body,
            startHead = _this$attributes2.startHead,
            endHead = _this$attributes2.endHead,
            startHeadOffset = _this$attributes2.startHeadOffset,
            endHeadOffset = _this$attributes2.endHeadOffset,
            rest = _objectWithoutPropertiesLoose(_this$attributes2, _excluded2); // append new arrow head


        this.appendArrowHead(this.getArrowHeadType(newValue), isStart);
        this.applyArrowStyle(rest, [isStart ? this.startHead : this.endHead]);
      }
    } else if (name === 'body') {
      // eslint-disable-next-line @typescript-eslint/no-unused-vars
      var _this$attributes3 = this.attributes,
          _body = _this$attributes3.body,
          _startHead = _this$attributes3.startHead,
          _endHead = _this$attributes3.endHead,
          _startHeadOffset = _this$attributes3.startHeadOffset,
          _endHeadOffset = _this$attributes3.endHeadOffset,
          _rest = _objectWithoutPropertiesLoose(_this$attributes3, _excluded3);

      this.removeChild(this.body, true); // @ts-ignore

      this.body = newValue;
      this.appendChild(this.body);
      this.applyArrowStyle(_rest, [this.body]);
    } else if (name === 'startHeadOffset') {
      this.moveArrowHeadAlongTangent(newValue, true);
    } else if (name === 'endHeadOffset') {
      this.moveArrowHeadAlongTangent(newValue, false);
    }
  };

  _proto.getArrowHeadType = function getArrowHeadType(head) {
    if (typeof head === 'boolean') {
      return 'default';
    }

    return 'custom';
  };

  _proto.appendArrowHead = function appendArrowHead(type, isStart) {
    var head;

    if (type === 'default') {
      head = this.createDefaultArrowHead();
    } else {
      // @ts-ignore
      head = isStart ? this.attributes.startHead : this.attributes.endHead;
    } // set position & rotation


    this.transformArrowHead(head, isStart); // heads should display on top of body

    head.setAttribute('zIndex', 1);

    if (isStart) {
      this.startHead = head;
    } else {
      this.endHead = head;
    }

    this.appendChild(head);
    var offset = isStart ? this.attributes.startHeadOffset : this.attributes.endHeadOffset;

    if (offset) {
      this.moveArrowHeadAlongTangent(offset, isStart);
    }
  }
  /**
   * transform arrow head according to arrow line
   */
  ;

  _proto.transformArrowHead = function transformArrowHead(head, isStart) {
    var position = create();
    var rad = 0;
    var x1 = 0;
    var x2 = 0;
    var y1 = 0;
    var y2 = 0;
    var bodyType = this.body && this.body.nodeName;

    if (bodyType === Shape.LINE) {
      var _this$body$attributes = this.body.attributes,
          _x1 = _this$body$attributes.x1,
          _x2 = _this$body$attributes.x2,
          _y1 = _this$body$attributes.y1,
          _y2 = _this$body$attributes.y2;
      x1 = isStart ? _x2 : _x1;
      x2 = isStart ? _x1 : _x2;
      y1 = isStart ? _y2 : _y1;
      y2 = isStart ? _y1 : _y2;
    } else if (bodyType === Shape.POLYLINE) {
      var points = this.body.attributes.points;
      var length = points.length;
      x1 = isStart ? points[1][0] : points[length - 2][0];
      y1 = isStart ? points[1][1] : points[length - 2][1];
      x2 = isStart ? points[0][0] : points[length - 1][0];
      y2 = isStart ? points[0][1] : points[length - 1][1];
    } else if (bodyType === Shape.PATH) {
      var _this$getTangent = this.getTangent(this.body, isStart),
          p1 = _this$getTangent[0],
          p2 = _this$getTangent[1];

      x1 = p1[0];
      y1 = p1[1];
      x2 = p2[0];
      y2 = p2[1];
    }

    var x = x1 - x2;
    var y = y1 - y2;
    rad = Math.atan2(y, x);
    position = fromValues(x2, y2, 0);

    if (isStart) {
      this.startHeadPosition = position;
      this.startHeadRad = rad;
    } else {
      this.endHeadPosition = position;
      this.endHeadRad = rad;
    }

    head.setLocalPosition(position);
    head.setLocalEulerAngles(rad * 180 / Math.PI + head.getLocalEulerAngles());
  };

  _proto.moveArrowHeadAlongTangent = function moveArrowHeadAlongTangent(offset, isStart) {
    var head = isStart ? this.startHead : this.endHead;

    if (head) {
      head.setLocalPosition(sub(create(), isStart ? this.startHeadPosition : this.endHeadPosition, fromValues(Math.cos(isStart ? this.startHeadRad : this.endHeadRad) * offset, Math.sin(isStart ? this.startHeadRad : this.endHeadRad) * offset, 0)));
    } // cut body


    if (this.body) ;
  };

  _proto.destroyArrowHead = function destroyArrowHead(isStart) {
    if (isStart && this.startHead) {
      // @ts-ignore
      this.removeChild(this.startHead, true);
      this.startHead = undefined;
    }

    if (!isStart && this.endHead) {
      // @ts-ignore
      this.removeChild(this.endHead, true);
      this.endHead = undefined;
    }
  };

  _proto.getTangent = function getTangent(path, isStart) {
    return isStart ? path.getStartTangent() : path.getEndTangent();
  };

  _proto.createDefaultArrowHead = function createDefaultArrowHead() {
    var _this$attributes4 = this.attributes,
        stroke = _this$attributes4.stroke,
        lineWidth = _this$attributes4.lineWidth;
    var sin = Math.sin,
        cos = Math.cos,
        PI = Math.PI;
    return new Path({
      style: {
        // draw an angle '<'
        path: "M" + 10 * cos(PI / 6) + "," + 10 * sin(PI / 6) + " L0,0 L" + 10 * cos(PI / 6) + ",-" + 10 * sin(PI / 6),
        stroke: stroke,
        lineWidth: lineWidth,
        transformOrigin: 'center',
        anchor: [0.5, 0.5] // set anchor to center

      }
    });
  };

  _proto.applyArrowStyle = function applyArrowStyle(attributes, objects) {
    var opacity = attributes.opacity,
        stroke = attributes.stroke,
        strokeOpacity = attributes.strokeOpacity,
        lineWidth = attributes.lineWidth,
        increasedLineWidthForHitTesting = attributes.increasedLineWidthForHitTesting;
    objects.forEach(function (shape) {
      if (shape) {
        if (!isNil(opacity)) {
          shape.style.opacity = opacity;
        }

        if (!isNil(stroke)) {
          shape.style.stroke = stroke;
        }

        if (!isNil(strokeOpacity)) {
          shape.style.strokeOpacity = strokeOpacity;
        }

        if (!isNil(lineWidth)) {
          shape.style.lineWidth = lineWidth;
        }

        if (!isNil(increasedLineWidthForHitTesting)) {
          shape.style.increasedLineWidthForHitTesting = increasedLineWidthForHitTesting;
        }
      }
    });
  };

  return Arrow;
}(CustomElement);
Arrow.tag = 'arrow';

export { Arrow };
